@@ -1,6 +1,6 @@
 %%%-------------------------------------------------------------------
 %%% @author Evgeny Khramtsov <ekhramtsov@process-one.net>
-%%% @copyright (C) 2002-2020 ProcessOne, SARL. All Rights Reserved.
+%%% @copyright (C) 2002-2025 ProcessOne, SARL. All Rights Reserved.
 %%%
 %%% Licensed under the Apache License, Version 2.0 (the "License");
 %%% you may not use this file except in compliance with the License.
@@ -20,7 +20,7 @@
 %% API
 -export([start/0, stop/0]).
 -export([parse/2, parse/3, validate/2, fail/2, replace_macros/1]).
--export([format_error/1, format_error/2, format_ctx/1]).
+-export([format_error/1, format_error/2, format_ctx/1, include_files/3]).
 %% Simple types
 -export([pos_int/0, pos_int/1, non_neg_int/0, non_neg_int/1]).
 -export([int/0, int/2, number/1, octal/0]).
@@ -108,11 +108,11 @@ parse(Path0, Validators, Opts) ->
     Path = unicode:characters_to_binary(Path0),
     {Opts1, Opts2} = proplists:split(
 		       proplists:compact(Opts),
-		       [replace_macros, include_files, plain_as_atom]),
+		       [replace_macros, include_files, plain_as_atom, additional_macros]),
     Opts3 = lists:flatten(Opts1),
     try
 	Y1 = read_yaml(prep_path(Path), Opts3, []),
-	Y2 = replace_macros(Y1, Opts3),
+	Y2 = replace_macros(define_additional_macros(Y1, Opts3), Opts3),
 	Validators1 = maps:merge(Validators, validators(Opts3)),
 	{ok, (options(Validators1, Opts2))(Y2)}
     catch _:{?MODULE, Why, Ctx} ->
@@ -243,7 +243,7 @@ binary() ->
 binary(Regexp) ->
     binary(Regexp, []).
 
--spec binary(iodata(), [re:compile_option()]) -> validator(binary()).
+-spec binary(iodata(), re:compile_options()) -> validator(binary()).
 binary(Regexp, Opts) when is_list(Regexp) orelse is_binary(Regexp) ->
     fun(Val) ->
 	    Bin = to_binary(Val),
@@ -265,7 +265,7 @@ string() ->
 string(Regexp) ->
     string(Regexp, []).
 
--spec string(iodata(), [re:compile_option()]) -> validator(string()).
+-spec string(iodata(), re:compile_options()) -> validator(string()).
 string(Regexp, Opts) when is_list(Regexp) orelse is_binary(Regexp) ->
     fun(Val) ->
 	    Str = to_string(Val),
@@ -361,18 +361,19 @@ url() ->
 url(Schemes) ->
     fun(Val) ->
 	    URL = to_binary(Val),
-	    case http_uri:parse(to_string(URL)) of
-		{ok, {_, _, Host, _, _, _}} when Host == ""; Host == <<"">> ->
+	    case parse_uri(URL) of
+		{ok, _, Host, _} when Host == ""; Host == <<"">> ->
 		    fail({bad_url, empty_host, URL});
-		{ok, {_, _, _, Port, _, _}} when Port =< 0 orelse Port >= 65536 ->
+		{ok, _, _, Port} when Port /= undefined,
+				      Port =< 0 orelse Port >= 65536 ->
 		    fail({bad_url, bad_port, URL});
-		{ok, {Scheme, _, _, _, _, _}} when Schemes /= [] ->
+		{ok, Scheme, _, _} when Schemes /= [] ->
 		    case lists:member(Scheme, Schemes) of
 			true -> URL;
 			false ->
 			    fail({bad_url, {unsupported_scheme, Scheme}, URL})
 		    end;
-		{ok, _} ->
+		{ok, _, _, _} ->
 		    URL;
 		{error, Why} ->
 		    fail({bad_url, Why, URL})
@@ -450,7 +451,7 @@ timeout(Unit, Inf) ->
 re() ->
     re([]).
 
--spec re([re:compile_option()]) -> validator(re:mp()).
+-spec re(re:compile_options()) -> validator(re:mp()).
 re(Opts) ->
     fun(Val) ->
 	    Bin = to_binary(Val),
@@ -464,7 +465,7 @@ re(Opts) ->
 glob() ->
     glob([]).
 
--spec glob([re:compile_option()]) -> validator(re:mp()).
+-spec glob(re:compile_options()) -> validator(re:mp()).
 glob(Opts) ->
     fun(Val) ->
 	    S = to_string(Val),
@@ -887,6 +888,19 @@ replace_macros(Y, Opts) ->
 	    Y
     end.
 
+define_additional_macros(Yi, Opts) ->
+    CurrentMacros = lists:flatten(proplists:get_all_values(define_macro, Yi)),
+    AdditionalMacros = proplists:get_value(additional_macros, Opts, []),
+    NewMacros = lists:foldl(fun({Key, Value}, CMs) ->
+                                     lists:keystore(Key, 1, CMs, {Key, Value})
+                             end,
+                             CurrentMacros,
+                             AdditionalMacros),
+    case NewMacros of
+        [] -> proplists:delete(define_macro, Yi);
+        _ -> [{define_macro, NewMacros} | proplists:delete(define_macro, Yi)]
+    end.
+
 -spec check_duplicated_macros([macro()]) -> [macro()].
 check_duplicated_macros(Macros) ->
     lists:foldl(
@@ -920,7 +934,11 @@ replace_macro(V, {Name, Val}, Macros, Path) ->
 		false -> replace_macros(Val, Macros, [Name|Path])
 	    end;
 	_ ->
-	    V
+	    case is_binary(V) andalso is_binary(Val) of
+	        true -> re:replace(V, <<"@", Name/binary, "@">>, Val,
+                                   [{return, binary}, global]);
+	        false -> V
+	    end
     end;
 replace_macro(Val, _, _, _) ->
     Val.
@@ -1143,6 +1161,35 @@ parse_ip_netmask(S) ->
 	    error
     end.
 
+-spec parse_uri(term()) -> {ok, atom(), binary(), integer() | undefined} |
+			   {error, term()}.
+-ifdef(USE_OLD_HTTP_URI).
+parse_uri(URL) when is_binary(URL) ->
+    parse_uri(to_string(URL)); % OTP =< 19's http_uri:parse/1 expects strings.
+parse_uri(URL) ->
+    case http_uri:parse(URL) of
+	{ok, {Scheme, _UserInfo, Host, Port0, _Path, _Query}} ->
+	    Port = if is_integer(Port0) -> Port0;
+		      true -> undefined
+		   end,
+	    {ok, Scheme, Host, Port};
+	{error, Reason} ->
+	    {error, Reason}
+    end.
+-else.
+parse_uri(URL0) ->
+    URL = re:replace(URL0, <<"@[A-Z]+@">>, <<"MACRO">>, [{return, binary}]),
+    case uri_string:parse(URL) of
+	URIMap when is_map(URIMap) ->
+	    Scheme = maps:get(scheme, URIMap, <<>>),
+	    Host = maps:get(host, URIMap, <<>>),
+	    Port = maps:get(port, URIMap, undefined),
+	    {ok, to_atom(Scheme), Host, Port};
+	{error, Reason, _Info} ->
+	    {error, Reason}
+    end.
+-endif.
+
 -spec fail(error_reason()) -> no_return().
 fail(Reason) ->
     fail(?MODULE, Reason).
